import argparse
import json
import sys
from contextlib import suppress
from typing import Any, Callable, Optional, cast

from m.core import Issue, OneOf
from m.core.ci_tools import error_block, get_ci_tool

from ..core.io import env
from .engine.misc import params_count
from .engine.sys import get_cli_command_modules
from .engine.types import CmdMap, CommandModule, MetaMap, NestedCmdMap
from .validators import validate_non_empty_str


def _main_parser(
    mod: NestedCmdMap,
    meta: MetaMap,
    add_args: Callable[[argparse.ArgumentParser], None] | None = None,
) -> argparse.Namespace:
    main_meta = meta['_root'].meta
    raw = argparse.RawTextHelpFormatter
    # NOTE: In the future we will need to extend from this class to be able to
    #   override the error method to be able to print CI environment messages.
    argp = argparse.ArgumentParser(
        formatter_class=raw,
        description=main_meta['description'] if main_meta else '...',
    )
    if add_args:
        add_args(argp)
    subp = argp.add_subparsers(
        title='commands',
        dest='command_name',
        required=True,
        help='additional help',
        metavar='<command>',
    )
    names = sorted(mod.keys())
    for name in names:
        if isinstance(mod[name], dict):
            meta_mod = meta[name]
            meta_dict = meta_mod.meta
            parser = subp.add_parser(
                name,
                help=meta_dict['help'],
                formatter_class=raw,
                description=meta_dict['description'],
            )
            if meta_mod.add_arguments:
                meta_mod.add_arguments(parser)
            subsubp = parser.add_subparsers(
                title='commands',
                dest='subcommand_name',
                required=True,
                help='additional help',
                metavar='<command>',
            )
            sub_mod = cast(CmdMap, mod[name])
            for subname in sorted(sub_mod.keys()):
                run_func = sub_mod[subname].run
                if params_count(run_func) == 2:
                    run_func(None, subsubp)
        else:
            mod_inst = cast(CommandModule, mod[name])
            run_func = mod_inst.run
            if params_count(run_func) == 2:
                run_func(None, subp)
    return argp.parse_args()


def run_cli(
    file_path: str,
    add_args: Callable[[argparse.ArgumentParser], None] | None = None,
) -> None:
    """Run the cli application.

    usage::

        def add_args(argp):
            argp.add_argument(...)
        def main():
            run_cli(__file__, add_args)

    We only need `add_args` if we need to gain access to the
    `argparse.ArgumentParser` instance.

    Args:
        file_path: The full name of the file: __file__.
        add_args: Optional callback to gain access to the ArgumentParser.
    """
    mod, meta = get_cli_command_modules(file_path)
    arg = _main_parser(mod, meta, add_args)

    run_func = None

    if hasattr(arg, 'subcommand_name'):
        sub_mod = cast(CmdMap, mod[arg.command_name])
        run_func = sub_mod[arg.subcommand_name].run
    else:
        run_func = cast(CommandModule, mod[arg.command_name]).run

    len_run_args = params_count(run_func)
    run_args = [arg, None][:len_run_args]
    # mypy is having a hard time figuring out the type of run_args
    exit_code = run_func(*run_args)  # type:ignore[arg-type]
    sys.exit(exit_code)


def display_issue(issue: Issue) -> None:
    """Print an error message.

    The error message is customize to the CI environment.

    Args:
        issue: An instance of an Issue.
    """
    get_ci_tool().error(issue.message)
    error_block(str(issue))


def display_result(inst: Any) -> None:
    """Print the JSON stringification of a non-None parameter.

    Args:
        inst: The instance to stringify.
    """
    if inst is not None:
        msg = inst
        with suppress(Exception):
            msg = json.dumps(inst, separators=(',', ':'))
        print(msg)  # noqa: WPS421


def run_main(
    callback: Callable[[], OneOf[Issue, Any]],
    handle_result: Callable[[Any], None] = display_result,
    handle_issue: Callable[[Issue], None] = display_issue,
) -> int:
    """Run the callback and print the returned value.

    To change how the result or an issue should be display provide the optional
    arguments `handle_result` and `handle_issue`. For instance, to display the
    raw value provide the `print` function.

    Args:
        callback: A function that returns a `OneOf`.
        handle_result: A function that takes in the Good result.
        handle_issue: A function that takes in the Issue.

    Returns:
        0 if the callback is a `Good` result otherwise return 1.
    """
    try:
        res = callback()
        result_value = res.value
        if res.is_bad:
            if isinstance(result_value, Issue):
                handle_issue(result_value)
            else:
                issue = Issue(
                    'non-issue exception',
                    cause=cast(Issue, result_value),
                )
                handle_issue(issue)
            return 1
        handle_result(result_value)
    except Exception as ex:
        issue = Issue('unknown caught exception', cause=ex)
        handle_issue(issue)
        return 1
    return 0


def error(msg: str, issue: Optional[Issue] = None) -> int:
    """Print an error message.

    Args:
        msg: The error message.
        issue: Optional `Issue` instance to go along with the error message.

    Returns:
        The integer 1.
    """
    get_ci_tool().error(msg)
    if issue:
        error_block(str(issue), sys.stderr)
    return 1


def cli_integration_token(
    integration: str,
    env_var: str,
) -> Callable[[argparse.ArgumentParser], argparse.Action]:
    """Return a function that takes in a parser.

    This generated function registers a token argument in the parser
    which looks for its value in the environment variables.

    Args:
        integration: The name of the integration.
        env_var: The environment variable name.

    Returns:
        A function that
    """
    return lambda parser: parser.add_argument(
        '-t',
        '--token',
        type=validate_non_empty_str,
        default=env(env_var),
        help=f'{integration} access token (default: env.{env_var})',
    )



import sys
from typing import TextIO

from m.core import Good, Issue, OneOf

from ..types import EnvVars, Message, ProviderModule

REPLACEMENTS = (
    ('|', '||'),
    ("'", "|'"),
    ('[', '|['),
    (']', '|]'),
    ('\n', '|n'),
)


# Do not use outside of module. This is done here to avoid
# disabling a flake8 for every ci tool method.
_print = print


def escape_msg(msg: str) -> str:
    """Escapes characters so Teamcity can print correctly.

    https://www.jetbrains.com/help/teamcity/build-script-interaction-with-teamcity.html#BuildScriptInteractionwithTeamCity-Escapedvalues

    Args:
        msg: The message to process.

    Returns:
        A message that can be provided to Teamcity.
    """
    message = msg
    for target, replacement in REPLACEMENTS:
        message = message.replace(target, replacement)
    return message


def _escape(name: str, description: str) -> str:
    desc = escape_msg(description)
    return f"{name}='{desc}'"


def _tc(msg: str, **kwargs) -> str:
    key_items = ' '.join(
        [_escape(name, desc) for name, desc in kwargs.items()],
    ).strip()
    return f'##teamcity[{msg} {key_items}]'


def env_vars() -> OneOf[Issue, EnvVars]:
    """Read basic environment variables from Teamcity.

    Returns:
        An EnvVar instance.
    """
    # WIP: Need to map other variables here. Not willing to do
    # at the moment since I'm using Github.
    return Good(EnvVars(ci_env=True))


def open_block(
    name: str,
    description: str,
    stream: TextIO | None = None,
) -> None:
    """Display the name and description of a block.

    https://www.jetbrains.com/help/teamcity/build-script-interaction-with-teamcity.html#BuildScriptInteractionwithTeamCity-BlocksofServiceMessages

    Args:
        name: Name of the block.
        description: The block description.
        stream: Defaults to sys.stdout.
    """
    _print(
        _tc('blockOpened', name=name, description=description),
        file=stream or sys.stdout,
    )


def close_block(name: str, stream: TextIO | None = None) -> None:
    """Close a previously defined block.

    Args:
        name: The name of the block to close.
        stream: Defaults to sys.stdout.
    """
    _print(_tc('blockClosed', name=name), file=stream or sys.stdout)


def error(msg: Message | str) -> None:
    """Print an error message.

    Doing so will make Teamcity abort the job.
    Supports file, line and col properties of Message.

    Args:
        msg: The message to display.
    """
    if isinstance(msg, str):
        _print(
            _tc('buildProblem', description=msg),
            file=sys.stderr,
        )
        return
    stream = msg.stream or sys.stderr
    parts = [x for x in (msg.file, msg.line, msg.col) if x]
    loc = ':'.join(parts)
    msg_info = f'[{loc}]: ' if loc else ''
    desc = f'{msg_info}{msg.message}' or ''
    _print(
        _tc('buildProblem', description=desc),
        file=stream,
    )


def warn(msg: Message | str) -> None:
    """Print an warning message to Teamcity.

    Supports file, line and col properties of Message.

    Args:
        msg: The message to display.
    """
    if isinstance(msg, str):
        _print(
            _tc('message', status='WARNING', text=msg),
            file=sys.stderr,
        )
        return
    stream = msg.stream or sys.stderr
    parts = [x for x in (msg.file, msg.line, msg.col) if x]
    loc = ':'.join(parts)
    msg_info = f'[{loc}]: ' if loc else ''
    desc = f'{msg_info}{msg.message}' or ''
    _print(
        _tc('message', status='WARNING', text=desc),
        file=stream,
    )


tool = ProviderModule(
    env_vars=env_vars,
    open_block=open_block,
    close_block=close_block,
    error=error,
    warn=warn,
)



from .ci_tools import error_block, get_ci_tool, warn_block
from .types import EnvVars, Message

# using as barrel file to export convenience functions
__all__ = (  # noqa: WPS410
    'get_ci_tool',
    'error_block',
    'warn_block',
    'Message',
    'EnvVars',
)



import inspect
import json
import sys
import traceback
from collections import OrderedDict
from contextlib import suppress
from typing import cast

from typing_extensions import TypedDict

IssueDict = TypedDict(
    'IssueDict',
    {
        'message': str,
        'description': str,
        'cause': object,
        'context': object,
        'traceback': list[str],
    },
    total=False,
)


def remove_traceback(issue_dict: object) -> None:
    """Remove the `traceback` key from a dictionary if it exists.

    It will recursively remove the `traceback` from its cause.

    Args:
        issue_dict: A dictionary representation of an `Issue`.
    """
    if isinstance(issue_dict, dict):
        if 'traceback' in issue_dict:
            issue_dict.pop('traceback')
        cause = issue_dict.get('cause')
        if cause:
            remove_traceback(cause)


# Rule WPS230 is suppressed here since I want to have a flat structure
# with this class. In other places we can refactor and group attributes into
# other classes.
class Issue(Exception):  # noqa: N818, WPS230 - Intention is not to raise
    """Wrapper to keep track of all exceptions.

    It provides a 'cause' field so that we may know why an issue was
    triggered.
    """

    show_traceback = True

    message: str
    description: str | None
    cause: Exception | None
    context: object | None
    include_traceback: bool
    cause_tb: list[str] | None

    def __init__(  # noqa: WPS211 - need to initialize the attributes
        self,
        message: str,
        description: str | None = None,
        cause: Exception | None = None,
        context: object | None = None,
        include_traceback: bool = True,
    ):
        """Create an Issue.

        Args:
            message: Simple description of the issue.
            description: More in depth detail on the issue
            cause: The exception/Issue that is responsible for this instance.
            context: Dictionary with any useful data related to the issue.
            include_traceback: If False, it skips computing the traceback.
        """
        super().__init__()
        self.message = message
        self.description = description
        self.cause = cause
        if cause and not isinstance(cause, Issue):
            # https://stackoverflow.com/a/12539332/788553
            with suppress(BaseException):
                exception_list = [
                    *traceback.format_tb(sys.exc_info()[2]),
                    *traceback.format_exception_only(
                        sys.exc_info()[0],
                        sys.exc_info()[1],
                    ),
                ]
                self.cause_tb = [
                    y
                    for x in exception_list
                    for y in x.splitlines()
                ]
        self.context = context
        self.include_traceback = include_traceback
        if self.include_traceback:
            frame = inspect.currentframe()
            self.traceback = [
                y
                for x in traceback.format_stack(frame)[:-1]
                for y in x.splitlines()
            ]

    def to_dict(self) -> IssueDict:
        """Convert to a ordered dictionary.

        This is done so that each of the properties are written in an expected
        order.

        Returns:
            An `IssueDict` instance.
        """
        issue_dict = cast(IssueDict, OrderedDict())
        issue_dict['message'] = self.message
        if self.description:
            issue_dict['description'] = self.description
        if self.context:
            issue_dict['context'] = self.context
        if self.include_traceback:
            issue_dict['traceback'] = self.traceback
        if self.cause:
            if isinstance(self.cause, Issue):
                issue_dict['cause'] = self.cause.to_dict()
            else:
                issue_dict['cause'] = {
                    'message': str(self.cause),
                    'traceback': self.cause_tb,
                }
        return issue_dict

    def to_str(self, show_traceback: bool) -> str:
        """Convert the instance to string.

        Args:
            show_traceback: If false, it will remove the error traceback.

        Returns:
            A string representation of the Issue instance.
        """
        issue_dict = self.to_dict()
        if not show_traceback:
            remove_traceback(issue_dict)
        return json.dumps(issue_dict, indent=2)

    def __str__(self) -> str:
        """Convert the instance to a string.

        Returns:
            A string representation of the Issue instance.
        """
        return self.to_str(Issue.show_traceback)



from m.cli import command
from pydantic import BaseModel, Field


class Arguments(BaseModel):
    """Close and immediately open another block."""

    to_close: str = Field(
        description='block name to close',
        positional=True,
        required=True,
    )
    name: str = Field(
        description='block name to open',
        positional=True,
        required=True,
    )
    description: str = Field(
        description='block description',
        positional=True,
        required=True,
    )


@command(
    name='sibling_block',
    help='close and open a sibling block',
    model=Arguments,
)
def run(arg: Arguments):
    from m.core.ci_tools import get_ci_tool
    tool = get_ci_tool()
    tool.close_block(arg.to_close)
    tool.open_block(arg.name, arg.description)
    return 0



from .engine.argparse import command
from .engine.types import FuncArgs, add_arg
from .utils import run_cli, run_main
from .validators import validate_json_payload, validate_payload

# using as barrel file to export convenience functions
__all__ = (  # noqa: WPS410
    'command',
    'add_arg',
    'validate_json_payload',
    'validate_payload',
    'run_main',
    'run_cli',
    'FuncArgs',
)



import math
import os
import sys
from typing import TypeVar

from m.core import Good, Issue, OneOf, issue
from pydantic import BaseModel

BaseModelT = TypeVar('BaseModelT', bound=BaseModel)


def format_seconds(number_of_seconds: int | float) -> str:
    """Return a string representing the number of seconds.

    The format is Xd:Xh:Xm:Xs:Xms.

    Args:
        number_of_seconds: The number of seconds to format.

    Returns:
        A friendly representation of the number of seconds.
    """
    milliseconds = int(math.floor(number_of_seconds * 1000))
    milli_sec = milliseconds % 1000
    seconds = int(math.floor(milliseconds / 1000))
    sec = seconds % 60
    minutes = int(math.floor(seconds / 60))
    mins = minutes % 60
    hours = int(math.floor(minutes / 60))
    hrs = hours % 24
    days = int(math.floor(hours / 24))

    entries = [
        f'{days}d' if days else '',
        f'{hrs}h' if hours else '',
        f'{mins}m' if mins else '',
        f'{sec}s' if sec else '',
        f'{milli_sec}ms' if milli_sec else '',
    ]
    return ':'.join([x for x in entries if x]) or '0s'


def env(name: str, def_val: str = '') -> str:
    """Access an environment variable.

    Args:
        name: The name of the environment variable.
        def_val: The default value to return if not set. Defaults to empty.

    Returns:
        The value of the environment variable if set otherwise the `def_val`.
    """
    return os.environ.get(name, def_val)


def renv(key: str) -> OneOf[Issue, str]:
    """Require an environment variable to be defined.

    Args:
        key: The environment variable required to be defined.

    Returns:
        A `OneOf` with the value of the environment variable or an issue.
    """
    env_value = os.environ.get(key)
    # Value may still be an empty string, checking against None
    if env_value is not None:
        return Good(env_value)
    return issue(f'missing {key} in env')


def env_model(model: type[BaseModelT]) -> OneOf[Issue, BaseModelT]:
    """Require multiple env vars to be defined.

    This can be done by defining a model::

        class GithubEnvVars(BaseModel):
            repo: str = Field('GITHUB_REPOSITORY')
            run_id: str = Field('GITHUB_RUN_ID')

    Then we use it::

        print(env.repo for env in env_model(GithubEnvVars)])

    Args:
        model: A pydantic model specifying the environment variables to fetch.

    Returns:
        A `OneOf` with the values of the environment variables or an issue.
    """
    schema = model.schema()
    missing: list[str] = []
    env_values: dict[str, str] = {}
    for name, field in schema['properties'].items():
        env_name = field.get('default')
        env_value = os.environ.get(env_name)
        if env_value is None:
            missing.append(env_name)
        else:
            env_values[name] = env_value
    if missing:
        missing_str = ', '.join(missing)
        return issue(f'missing [{missing_str}] in env')
    return Good(model(**env_values))


def _ver_str(major: int, minor: int, patch: int) -> str:
    return f'{major}.{minor}.{patch}'


def prompt_next_version(version: str, release_type: str) -> str:
    """Prompt the developer to select the next version.

    It displays the possible major, minor and patch versions and prompts the
    developer to enter one of them. See https://semver.org/.

    Args:
        version: The current version.
        release_type: If `hotfix` then it bumps the patch value.

    Returns:
        One of the versions.
    """
    ver = version.split('-')[0]
    parts = [int(x) for x in ver.split('.')]
    patch = _ver_str(parts[0], parts[1], parts[2] + 1)
    if release_type == 'hotfix':
        return patch

    minor = _ver_str(parts[0], parts[1] + 1, 0)
    major = _ver_str(parts[0] + 1, 0, 0)
    valid = False
    options = [minor, major]
    msg = f'Current version is {version}. Enter one of the following:\n  '
    msg += '\n  '.join(options)
    dev_input = ''
    while not valid:
        print(msg, file=sys.stderr)
        dev_input = input('')
        if dev_input in options:
            valid = True
    return dev_input



from m.cli import command, run_main
from pydantic import BaseModel


class Arguments(BaseModel):
    """Display a single word representing the current git status.

    example::

        $ m git status
        clean

    statuses::

        unknown
        untracked
        stash
        clean
        ahead
        behind
        staged
        dirty
        diverged
        ?
    """


@command(
    name='status',
    help='display the current git status',
    model=Arguments,
)
def run():
    from m import git
    return run_main(git.get_status, print)



import sys
from typing import TextIO

from m.core.io import env

from .providers.github import tool as gh_tool
from .providers.local import tool as local_tool
from .providers.teamcity import tool as tc_tool
from .types import ProviderModule

# Do not use outside of module. This is done here to avoid
# disabling a flake8 for every ci tool method.
_print = print


def get_ci_tool() -> ProviderModule:
    """Return the current CI Tool based on the environment variables.

    Returns:
        A `ProviderModule` instance with methods to provide messages in
        a CI environment.
    """
    if env('GITHUB_ACTIONS'):
        return gh_tool
    if env('TC') or env('TEAMCITY'):
        return tc_tool
    return local_tool


def error_block(msg: str, stream: TextIO | None = None) -> None:
    """Print an issue within an error block.

    Args:
        msg: The error message to display.
        stream: The file stream to print (defaults to sys.stderr)
    """
    tool = get_ci_tool()
    tool.open_block('error', '', stream or sys.stderr)
    _print(msg, file=stream or sys.stderr)
    tool.close_block('error', stream or sys.stderr)


def warn_block(msg: str, stream: TextIO | None = None) -> None:
    """Print an issue within a warning block.

    Args:
        msg: The warning message to display.
        stream: The file stream to print (defaults to sys.stderr)
    """
    tool = get_ci_tool()
    tool.open_block('warning', '', stream or sys.stderr)
    _print(msg, file=stream or sys.stderr)
    tool.close_block('warning', stream or sys.stderr)



import sys
from typing import TextIO, cast

from m.core import Issue, OneOf, issue, one_of
from m.core.io import env_model
from pydantic import BaseModel, Field

from ..types import EnvVars, Message, ProviderModule

# Do not use outside of module. This is done here to avoid
# disabling a flake8 for every ci tool method.
_print = print


class GithubEnvVars(BaseModel):
    """Environment variables required when running in Github."""

    repo: str = Field('GITHUB_REPOSITORY')
    run_id: str = Field('GITHUB_RUN_ID')
    run_number: str = Field('GITHUB_RUN_NUMBER')
    github_token: str = Field('GITHUB_TOKEN')
    git_branch: str = Field('GITHUB_REF')
    git_sha: str = Field('GITHUB_SHA')
    triggered_by: str = Field('GITHUB_ACTOR')


def env_vars() -> OneOf[Issue, EnvVars]:
    """Read the environment variables from Github Actions.

    Returns:
        An `EnvVars` instance or an Issue.
    """
    server_url = 'https://github.com'
    return one_of(
        lambda: [
            EnvVars(
                **env.dict(),
                ci_env=True,
                server_url=server_url,
                run_url=run_url,
            )
            for env in env_model(GithubEnvVars)
            for run_url in (
                f'{server_url}/{env.repo}/actions/runs/{env.run_id}',
            )
        ],
    ).flat_map_bad(lambda x: issue(
        'GH Actions env_vars failure', cause=cast(Issue, x),
    ))


def open_block(
    name: str,
    description: str,
    stream: TextIO | None = None,
) -> None:
    """Group log lines.

    There is no description field and it does not support nesting.

    https://docs.github.com/en/actions/reference/workflow-commands-for-github-actions#grouping-log-lines

    Args:
        name: The name of the block to open.
        description: Not supported.
        stream: Defaults to sys.stdout.
    """
    # pylint: disable=unused-argument
    _print(f'::group::{name}', file=stream or sys.stdout)


def close_block(name: str, stream: TextIO | None = None) -> None:
    """Close the current block.

    Github Actions does not support nesting blocks, all this does is close
    the current block.

    Args:
        name: Has no effect.
        stream: Defaults to sys.stdout.
    """
    # pylint: disable=unused-argument
    _print('::endgroup::', file=stream or sys.stdout)


def error(msg: Message | str) -> None:
    """Print an error message.

    https://docs.github.com/en/actions/reference/workflow-commands-for-github-actions#setting-an-error-message

    Args:
        msg: The message to display.
    """
    if isinstance(msg, str):
        _print(f'::error::{msg}', file=sys.stderr)
        return
    stream = msg.stream or sys.stderr
    file_entry = f'file={msg.file}' if msg.file else ''
    line_entry = f'line={msg.line}' if msg.line else ''
    col_entry = f'col={msg.col}' if msg.col else ''
    parts = [x for x in (file_entry, line_entry, col_entry) if x]
    loc = ','.join(parts)
    msg_info = f' {loc}' if loc else ''
    _print(f'::error{msg_info}::{msg.message}', file=stream)


def warn(msg: Message | str) -> None:
    """Print an warning message.

    https://docs.github.com/en/actions/reference/workflow-commands-for-github-actions#setting-a-warning-message

    Args:
        msg: The message to display.
    """
    if isinstance(msg, str):
        _print(f'::warning::{msg}', file=sys.stderr)
        return
    stream = msg.stream or sys.stderr
    file_entry = f'file={msg.file}' if msg.file else ''
    line_entry = f'line={msg.line}' if msg.line else ''
    col_entry = f'col={msg.col}' if msg.col else ''
    parts = [x for x in (file_entry, line_entry, col_entry) if x]
    loc = ','.join(parts)
    msg_info = f' {loc}' if loc else ''
    _print(f'::warning{msg_info}::{msg.message}', file=stream)


tool = ProviderModule(
    env_vars=env_vars,
    open_block=open_block,
    close_block=close_block,
    error=error,
    warn=warn,
)



import sys
from typing import TextIO

from m.core import Issue, OneOf, one_of

from m import git

from ..types import EnvVars, Message, ProviderModule

# Do not use outside of module. This is done here to avoid
# disabling a flake8 for every ci tool method.
_print = print


def env_vars() -> OneOf[Issue, EnvVars]:
    """Obtain basic environment variables in a local environment.

    Returns:
        An `EnvVars` instance if successful.
    """
    return one_of(
        lambda: [
            EnvVars(
                git_branch=git_branch,
                git_sha=git_sha,
            )
            for git_branch in git.get_branch()
            for git_sha in git.get_current_commit_sha()
        ],
    )


def open_block(
    name: str,
    description: str,
    stream: TextIO | None = None,
) -> None:
    """Display the name and description of a block.

    Local environments do not support blocks like Github and Teamcity.

    Args:
        name: Name of the block.
        description: The block description.
        stream: Defaults to sys.stdout.
    """
    desc = f' {description}' if description else ''
    _print(f'{name}:{desc}', file=stream or sys.stdout)


def close_block(name: str, stream: TextIO | None = None) -> None:
    """Display an empty line to denote the end of a block.

    Args:
        name: The name of the block to close.
        stream: Defaults to sys.stdout.
    """
    # pylint: disable=unused-argument
    _print('', file=stream or sys.stdout)


def error(msg: Message | str) -> None:
    """Print an error message.

    Supports file, line and col properties of Message.

    Args:
        msg: The message to display.
    """
    if isinstance(msg, str):
        _print(f'error: {msg}', file=sys.stderr)
        return
    stream = msg.stream or sys.stderr
    parts = [x for x in (msg.file, msg.line, msg.col) if x]
    loc = ':'.join(parts)
    msg_info = f'[{loc}]' if loc else ''
    _print(f'error{msg_info}: {msg.message}', file=stream or sys.stderr)


def warn(msg: Message | str) -> None:
    """Print an warning message.

    Supports file, line and col properties of Message.

    Args:
        msg: The message to display.
    """
    if isinstance(msg, str):
        _print(f'warn: {msg}', file=sys.stderr)
        return
    stream = msg.stream or sys.stderr
    parts = [x for x in (msg.file, msg.line, msg.col) if x]
    loc = ':'.join(parts)
    msg_info = f'[{loc}]' if loc else ''
    _print(f'warn{msg_info}: {msg.message}', file=stream)


tool = ProviderModule(
    env_vars=env_vars,
    open_block=open_block,
    close_block=close_block,
    error=error,
    warn=warn,
)



import json
import sys
from collections.abc import Mapping
from contextlib import suppress
from pathlib import Path
from typing import Any, List
from typing import Mapping as Map
from typing import Union, cast

from .ci_tools import get_ci_tool
from .fp import Good, OneOf
from .issue import Issue
from .one_of import issue


def read_json(
    filename: str | None,
    error_if_empty: bool = False,
) -> OneOf[Issue, Any]:
    """Read a json object from a json file.

    Args:
        filename: The filename to read from, if `None` it reads from stdin.
        error_if_empty: The json parser may throw an error if `True`.

    Returns:
        A `Good` containing the parsed contents of the json file.
    """
    empty: str = '' if error_if_empty else 'null'
    try:
        if filename is None:
            return Good(json.loads(sys.stdin.read() or empty))
        with Path.open(Path(filename), encoding='UTF-8') as file_handle:
            return Good(json.loads(file_handle.read() or empty))
    except Exception as ex:
        return issue(
            'failed to read json file',
            context={'filename': filename or 'SYS.STDIN'},
            cause=ex,
        )


def parse_json(
    json_str: str,
    error_if_empty: bool = False,
) -> OneOf[Issue, Any]:
    """Parse a string as json.

    Args:
        json_str: The string to parse.
        error_if_empty: The json parser may throw an error if `True`.

    Returns:
        A `Good` containing the parsed contents of the json string.
    """
    try:
        empty = '' if error_if_empty else 'null'
        return Good(json.loads(json_str or empty))
    except Exception as ex:
        return issue('failed to parse the json data', cause=ex)


def get(dict_inst: Any, key_str: str) -> OneOf[Issue, Any]:
    """Return the value based on the `key_str` specified.

    The following are equivalent::

        obj['a']['b']['c'] <=> get(obj, 'a.b.c').value

    provided that the keys `a`, `b` and `c` are valid.

    Args:
        dict_inst: The dictionary instance to query.
        key_str: A simple query to fetch from the dict_inst.

    Returns:
        A `Good` if the value we want is available, otherwise it returns a
        `Bad` with the path that returned `None`.
    """
    keys = key_str.split('.')
    current = dict_inst
    for num, key in enumerate(keys):
        new_key: Union[str, int] = key
        with suppress(ValueError):
            new_key = int(key)
        try:
            current = current[new_key]
        except KeyError:
            pth = '.'.join(keys[:num + 1])
            return issue(f'`{pth}` path was not found')
        except Exception as ex:
            pth = '.'.join(keys[:num])
            if not isinstance(current, Mapping):
                context = pth or current
                return issue(f'`{context}` is not a dict')
            # catch unknown issue
            return issue(  # pragma: no cover
                f'{pth} resulted in an error',
                cause=ex,
            )
    return Good(current)


def multi_get(
    dict_inst: object,
    *keys: str,
) -> OneOf[Issue, List[Any]]:
    """Call `get` for every input specified by `keys`.

    It collects the invalid keys and returns an `Issue`::

        multi_get(obj, 'a', 'a.b', 'a.b.c')

    Args:
        dict_inst: The dictionary instance to query.
        keys: The queries to apply.

    Returns:
        A `Good` with a list of the results if successful, otherwise a `Bad`
        with the list of failures.
    """
    result_items = []
    failures = []
    for key in keys:
        res = get(dict_inst, key)
        if res.is_bad:
            failures.append(
                Issue(
                    message=f'key lookup failure: `{key}`',
                    cause=res.value,
                    include_traceback=False,
                ),
            )
        else:
            result_items.append(res.value)
    if failures:
        return issue(
            'multi_get key retrieval failure',
            context=[x.to_dict() for x in failures],
            include_traceback=False,
        )
    return Good(result_items)


def _to_str(inst: Any):
    if isinstance(inst, bool):
        return f'{inst}'.lower()
    if inst is None:
        return 'null'
    return f'{inst}'


def jsonq(
    dict_inst: Map[str, Any],
    separator: str,
    display_warning: bool,
    *key_str: str,
) -> int:
    """Print the values obtained from `multi_get` to stdout.

    Args:
        dict_inst: The dictionary instance to query.
        separator: A string to use to separate the results.
        display_warning: Print a warning if true.
            The process will still exit with 1 but the ci tool
            will not exit when it sees the error message.
        key_str: The queries to apply.

    Returns:
        0 if all the keys are available, non-zero if there are problems.
    """
    tool = get_ci_tool()
    either = multi_get(dict_inst, *key_str)
    if either.is_bad:
        problem = cast(Issue, either.value)
        if display_warning:
            tool.warn(problem.message)
        else:
            tool.error(problem.message)
        print(either.value, file=sys.stderr)  # noqa: WPS421
        return 1
    strings = [_to_str(res_item) for res_item in cast(List[Any], either.value)]
    print(separator.join(strings))  # noqa: WPS421 - meant to use on cli
    return 0
